from TiledLayer import *
from actors import *
from cocos.director import Director
from gamedata import *

class Level(TiledLayer):
    def __init__(self, tileset, width, height, pos_player=(100,100)):
        super(Level, self).__init__(tileset, width, height)
        pos = (pos_player[0] * self.tile_width , 
                self.height - pos_player[1] * self.tile_height)

        self.bg = None

        self.player = Player(pos)
        self.add(self.player, z = 2)
        self.bird = Bird(pos)
        self.add(self.bird, z = 2)

        # a camera precisa comecar focalizando o personagem
        self.need_focus = True

        self.unschedule(self.update)
        self.schedule_interval(self.update, 1.0/FPS)

    def set_bg(self, sp):
        x = sp.width/2
        y = sp.height/2
        sp.position = (x, y)
        self.add(sp, z=0)
        self.bg = sp

    def generate(self):
        super(Level, self).generate()
        #center player on level begin
        x,y = self.player.position
        vpx, vpy = (x - 320, y - 480)
        if (vpx <= 0): vpx = 0
        if (vpy <= 0): vpy = 0
        if (vpx + 640 >= self.width): vpx = self.width - 640
        if (vpy + 480 >= self.height): vpy = self.height - 480
        self.view_port = (vpx, vpy)
        
    def global_coords(self, pos):
        return self.view_port[0] + pos[0], self.view_port[1] + pos[1]
        
    def screen_coords(self, pos):
        return  pos[0] - self.view_port[0], pos[1] - self.view_port[1]

    def snap_to_grid(self, point):
        return (point[0], point[1])

    def collides(self, point):
        mx, my = self.get_index(point)
        val = self[my, mx]
        if val[0] == BLOCK:
            return True
        return False

    def get_tile(self, point):
        mx, my = self.get_index(point)
        return self[my, mx]
        
    def _scroll(self, dt):
        x, y = self.screen_coords(self.player.position)
        w, h = 640, 480

        # a camera precisa comecar focalizando o personagem
        if self.need_focus:
            self.need_focus = False
            (vx, vy) = (100,100)
        else:
            vy = abs(self.player.speed[1]*dt)
        if x >  w/2 + 10 and self.can_scroll_right(self.player.speed[0]*dt):
            self.move(self.player.speed[0]*dt, 0)
        elif x < w/2 - 10 and self.can_scroll_left(-self.player.speed[0]*dt):
            self.move(-self.player.speed[0]*dt, 0)
        if y >  h/2 + 10 and self.can_scroll_up(vy):
            self.move(0, vy)
        elif y < h/2 - 10 and self.can_scroll_down(-vy):
            self.move(0, -vy)
        
    def update(self, dt):
        if dt > 0.1: return
        self._scroll(dt)
        
