from cocos.sprite import *
from cocos.layer import ColorLayer
from gamedata import *
import pyglet
import data

from cocos.director import *
from cocos.scene import *

#quad_width = 640
quad_width = 1760
quad_height = 480

class TiledLayer(ColorLayer):
    def __init__(self, tileset, width, height):
        super(TiledLayer, self).__init__(0, 0, 100, 255)
        self._load_sprite_sheet(tileset, width, height)
        self.images_values = [ [None for x in range(width/self.tile_width)] 
                                for y in range(height/self.tile_height)]
        self.view_port = (0, 0)
        self.width = width
        self.height = height
        #print width, height
        
    def _create_quad(self):
        return Sprite(pyglet.image.create(quad_width, quad_height).get_texture())
        
    def _create_quads(self):
        self.bl = self._create_quad()
        self.br = self._create_quad()
        self.tl = self._create_quad()
        self.tr = self._create_quad()
        
    def _load_sprite_sheet(self, ssheet, width, height):
        img = pyglet.image.load( data.filepath(ssheet[0]) )
        self.tile_rows = ssheet[1]
        self.tile_cols = ssheet[2]
        self.sp_sheet = pyglet.image.ImageGrid(img, self.tile_rows, self.tile_cols)
        self.tile_width = img.width / self.tile_cols
        self.tile_height = img.height / self.tile_rows     
        
    def __getitem__(self, index):
        if type(index) == type((0,1)):
            i,j = index
            if i >= 0 and j >= 0 and i < len(self.images_values) and j < len(self.images_values[0]) and self.images_values[i][j]:
                val = self.images_values[i][j][1]
                return (val[0], val[1])
        return (BLOCK, None)
            
    def __setitem__(self, index, value):
        if type(index) == type((0,1)):
            i, j = index
            if i >= 0 and j >= 0 and i < len(self.images_values) and j < len(self.images_values[0]):
                self.images_values[i][j] = (value[0], (value[1], value[2]))
            else:
                #print "EROOO", i, j
                pass
        
    def get_index(self, point):
        mx = int(point[0]/self.tile_width)
        #print self.height/self.tile_height, self.height, self.tile_height
        my = int(self.height/self.tile_height) - 1 - int(point[1]/self.tile_height)
        return mx, int(my)
        
    def _find_points(self):
        bl = (int(self.view_port[0]), int(self.view_port[1]))
        br = (bl[0] + quad_width, bl[1])
        tl = (bl[0], bl[1] + quad_height)
        tr = (bl[0] + quad_width, bl[1] + quad_height)
        return bl, br, tl, tr
            
    def _get_quad(self, point):
        return (point[0]/quad_width, point[1]/quad_height)
        
    def _draw_quad(self, quad, quad_pos):
        quad_x, quad_y = quad_pos
        fromy0 = False
        from_x = quad_x * quad_width / self.tile_width
        to_x = min(from_x + quad_width / self.tile_width, self.width/self.tile_width)
        to_y = self.height/self.tile_height - quad_y * quad_height / self.tile_height
        if to_y - quad_height / self.tile_height < 0:
            fromy0 = True
        from_y = max(to_y - quad_height / self.tile_height, 0)
        for y in range(from_y, to_y):
            for x in range(from_x, to_x):
                pos_x = (x-from_x)*self.tile_width
                if fromy0:
                    pos_y = to_y*self.tile_height - (y-from_y+1)*self.tile_height
                else:
                    pos_y = quad_height - (y-from_y+1)*self.tile_height
                if self.images_values[y][x] != None:
                    img = self.sp_sheet[self.images_values[y][x][0]]
                    quad.image.blit_into(img, pos_x, pos_y, 0)
        quad.x = quad.x_initial = (quad_x + 1) * quad_width - quad_width/2
        quad.y = quad.y_initial = (quad_y + 1) * quad_height - quad_height/2
        
    def _position_quads(self):
        for quad in [self.bl, self.br, self.tl, self.tr]:
            quad.x = quad.x_initial - self.view_port[0]
            quad.y = quad.y_initial - self.view_port[1]
        
    def _gen_quad(self, point, quad):
        quad_t = self._get_quad(point)
        self._draw_quad(quad, quad_t)
        self.add(quad, z = 1)
        
    def generate(self):
        self.generate_quads()
        
    def generate_quads(self):
        self._create_quads()
        bl, br, tl, tr = self._find_points()
        self._gen_quad(bl, self.bl)
        self._gen_quad(br, self.br)
        self._gen_quad(tl, self.tl)
        self._gen_quad(tr, self.tr)
        self._position_quads()

    def _update_right(self):
        self.remove(self.bl)
        self.remove(self.tl)
        self.bl = self.br
        self.tl = self.tr
        bl, br, tl, tr = self._find_points()
        self.br = self._create_quad()
        self.tr = self._create_quad()
        self._gen_quad(br, self.br)
        self._gen_quad(tr, self.tr)
        
    def _update_left(self):
        self.remove(self.br)
        self.remove(self.tr)
        self.br = self.bl
        self.tr = self.tl
        bl, br, tl, tr = self._find_points()
        self.bl = self._create_quad()
        self.tl = self._create_quad()
        self._gen_quad(bl, self.bl)
        self._gen_quad(tl, self.tl)

    def _update_up(self):
        self.remove(self.br)
        self.remove(self.bl)
        self.bl = self.tl
        self.br = self.tr
        bl, br, tl, tr = self._find_points()
        self.tl = self._create_quad()
        self.tr = self._create_quad()
        self._gen_quad(tl, self.tl)
        self._gen_quad(tr, self.tr)

    def _update_down(self):
        self.remove(self.tr)
        self.remove(self.tl)
        self.tr = self.bl
        self.tl = self.br
        bl, br, tl, tr = self._find_points()
        self.bl = self._create_quad()
        self.br = self._create_quad()
        self._gen_quad(bl, self.bl)
        self._gen_quad(br, self.br)
        
    def can_scroll_left(self, dx):
        return self.view_port[0]+dx >= 0
        
    def can_scroll_right(self, dx):
        #x_limit = abs(self.width - quad_width)
        x_limit = abs(self.width - SCREEN_WIDTH)
        return self.view_port[0]+dx <= x_limit

    def can_scroll_down(self, dy):
        return self.view_port[1]+dy >= 0
        
    def can_scroll_up(self, dy):
        y_limit = abs(self.height - quad_height)
        return self.view_port[1]+dy <= y_limit
        
    def move(self, dx, dy):
        if not ((self.can_scroll_up(dy) and dy > 0) or (self.can_scroll_down(dy)
            and dy < 0)):
            dy = 0
        if not (self.can_scroll_left(dx) and self.can_scroll_right(dx)):
            dx = 0
        self.view_port = (self.view_port[0]+dx, self.view_port[1]+dy)
        if self.bl.x + quad_width/2 < 0:
            self._update_right()
        elif self.bl.x + quad_width/2 > quad_width:
            self._update_left()
        if self.bl.y + quad_height/2 < 0:
            self._update_up()
        elif self.bl.y + quad_height/2 > quad_height:
            self._update_down()
        self._position_quads()


if __name__ == '__main__':
    director.init()
    t = TiledLayer( ('tileset1.png', 1, 11), 2400, 480  )
    t[0,2] = (0, 2, 3)
    t.generate_quads()
    director.run(Scene( t ))

