# --------------------------------
# actors.py
# Modulo dos atores
# --------------------------------

from testlevel import *
from gameover import *
from congratulations import *
from cocos.director import *
from cocos.layer import *
from cocos.sprite import *
from cocos.actions import *
from gamedata import *
from pyglet.gl import *
import pyglet
import math
import random
import sound
import data




# Actor
# representa um ator generico
class Actor(Sprite):

    # __init__()
    # construtor
    def __init__(self, animation_name, position=(0,0), rotation=0, scale=1, opacity=255, color=(255,255,255), anchor=None, flip_x=False, flip_y=False, blend_mode = BLEND_STANDARD, hp=100, mp=100):
        self.spawn_point = position
        anim = animation[animation_name]
        img = animgrid[anim[0]][anim[1][0]]
        if anchor is None: anchor = (img.width/2, img.height)
        self.anchor = anchor
        super(Actor, self).__init__(img, position, rotation, scale, opacity, color, anchor)

        self._maxhp = hp
        self._maxmp = mp
        self._hp = hp
        self._mp = mp

        self.flip_x = flip_x
        self.flip_y = flip_y
        self._dead = False
        self.blend_mode = blend_mode;
        self.visible = True
        self.blink_timer = 0
        self.blink_aux = 0

        self.animation = None
        self.change_animation(animation_name)

    # _setup_blending()
    # configura blending
    def _setup_blending(self):
        if self.blend_mode == BLEND_ADDITIVE:
            glBlendFunc(GL_SRC_ALPHA, GL_ONE)
        else:
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    # draw()
    # substitui draw do sprite
    def draw(self):
        if self.visible:
            #converte para coordenadas da tela
            x, y = self.position
            self.position = self.parent.screen_coords((x,y))
            self._group.set_state()
            self._setup_blending()
            if self._vertex_list is not None:
                self._vertex_list.draw(GL_QUADS)
            self._group.unset_state()
            self.position = x, y

    # change_animation()
    # muda a animacao
    def change_animation(self, animation_name):
        self.animation_name = animation_name
        if self.flip_x: animation_name += '_hflip'
        if self.flip_y: animation_name += '_vflip'
        anim = animation[animation_name]
        if not(anim is self.animation):
            self.animation = anim
            self.animation_frame = 0.0

    # update()
    # atualiza ator
    def update(self, dt):
        self._myname = '%s_%s_%d_%d' % (self.parent.mapfile, self.animation_name, self.spawn_point[0], self.spawn_point[1])
        self.animation_frame += self.animation[2]*dt
        if self.animation_frame >= len(self.animation[1]):
            if self.animation[3]: # loop?
                self.animation_frame = 0.0
            else:
                self.animation_frame = len(self.animation[1])-1
        self.image = animgrid[self.animation[0]][self.animation[1][int(self.animation_frame)]]
        self.image.anchor_x = self.anchor[0]
        self.image.anchor_y = self.anchor[1]
        self.change_animation( self.animation_name ) # flip bugfix

        # blinking
        self.blink_timer = max(0, self.blink_timer-dt)
        self.blink_aux += dt
        if self.blink_aux > 0.1:
            self.blink_aux = 0
            if self.is_blinking():
                self.visible = not self.visible
            else:
                self.visible = True

        # nao volte nunca mais
        if self._myname in entity_set:
            self.destroy()

    # destroy()
    # tira ator da cena
    def destroy(self):
        self._dead = True

    # destroy_forever()
    # tira este ator da cena e faz com
    # que ele nunca mais volte
    def destroy_forever(self):
        global entity_set
        entity_set.add(self._myname)
        self.destroy()

    # is_dead()
    # ator morto?
    def is_dead(self):
        return self._dead

    # kill()
    # mata este actor
    def kill(self, speed=None):
        if speed is None: speed = (-self.speed[0], -self.speed[1])
        if not(self.is_dead()):
            self.parent.create_actor( DeadActor(self, speed), zindex=2 )
            self.destroy_forever()
            sound.play_sample('Chip084.wav')

    # collides()
    # testa colisao entre self e act
    def collides(self, act):
        #(sx,sy) = (self.position[0]-self.anchor[0], self.position[1]-self.anchor[1])
        (sx,sy) = self.position
        (sw,sh) = (self.image.width, self.image.height)
        #(dx,dy) = (act.position[0]-act.anchor[0], act.position[1]-act.anchor[1])
        (dx,dy) = act.position
        (dw,dh) = (act.image.width, act.image.height)
        a = (sx,sy,sx+sw,sy+sh)
        b = (dx,dy,dx+dw,dy+dh)
        return (a[0]<b[2] and a[2]>b[0] and a[1]<b[3] and a[3]>b[1])

    # get_collision_detectors()
    # retorna uma octupla indicando se os tiles nas
    # coordenadas topleft, top, topright, right, bottomright,
    # bottom, bottomleft, left sao obstaculos
    # tupla da forma: (TL, T, TR, R, BR, B, BL, L)
    def get_collision_detectors(self, epsilon=(2,0)):
        (x,y) = (self.position[0]-self.anchor[0], self.position[1]-self.anchor[1])
        (w,h) = (self.image.width, self.image.height)
        (ex, ey) = epsilon

        tl = (x-ex, y+h+ey)
        t = (x+w/2, y+h+ey)
        tr = (x+w+ex, y+h+ey)
        r = (x+w+ex, y+h/2)
        br = (x+w+ex, y-ey) 
        b = (x+w/2, y-ey)
        bl = (x-ex, y-ey)
        l = (x-ex, y+h/2)

        f = lambda k: self.parent.collides(k)
        return (f(tl), f(t), f(tr), f(r), f(br), f(b), f(bl), f(l))

    # get_my_tile()
    # retorna o tile que fica na minha posicao
    def get_my_tile(self):
        (x,y) = (self.position[0]-self.anchor[0], self.position[1]-self.anchor[1])
        (w,h) = (self.image.width, self.image.height)
        p = (x+w/2, y+h/2)
        return self.parent.get_tile(p)


    # is_blinking()
    # este actor esta piscando?
    def is_blinking(self):
        return self.blink_timer > 0

    # blink()
    # pisca durante alguns segundos
    def blink(self, seconds):
        if not(self.is_blinking()):
            self.blink_timer = seconds

    # set_hp()
    # define a qtd de hp (sem piscar)
    def set_hp(self, x):
        self._hp = max(0, min(self._maxhp, x))
        if self._hp == 0:
            self.kill()

    # get_hp()
    # devolve a qtd de hp
    def get_hp(self):
        return self._hp

    # get_maxhp()
    # devolve a qtd maxima de hp
    def get_maxhp(self):
        return self._maxhp

    # set_mp()
    # define a qtd de mp
    def set_mp(self, x):
        self._mp = max(0, min(self._maxmp, x))

    # get_mp()
    # devolve a qtd de mp
    def get_mp(self):
        return self._mp

    # get_maxmp()
    # devolve a qtd maxima de mp
    def get_maxmp(self):
        return self._maxmp

    # hit()
    # causa dano ao actor
    def hit(self, damage):
        if not(self.is_blinking()):
            self.set_hp( self.get_hp()-damage )
            self.blink(BLINK_DURATION)
            if self.get_hp() > 0:
                sound.play_sample('Chip047.wav')

    # inside_screen()
    # o actor esta dentro da tela visivel?
    def inside_screen(self):
        eps = (-self.image.width, -self.image.height)
        (x,y) = self.parent.screen_coords(self.position)
        return not((x-eps[0] < 0) or (x+eps[0] >= SCREEN_WIDTH) or (y-eps[1] < 0) or (y+eps[1] >= SCREEN_HEIGHT))


# Collectable
# item coletavel
class Collectable(Actor):

    # __init__()
    # construtor
    def __init__(self, animation_name, position, **kwargs):
        super(Collectable, self).__init__(animation_name, position, **kwargs)
        self.objenabled = True

    # on_touch()
    # este item foi coletado pelo player. Funcao virtual.
    def on_touch(self):
        pass

    # update()
    # atualizacao
    def update(self, dt):
        super(Collectable, self).update(dt)
        self.image.anchor = self.anchor = (-self.image.width/2, self.image.height)

        # aaargh, o player me pegou!
        if self.collides( self.parent.player ) and self.objenabled:
            self.objenabled = False
            self.on_touch()

        # estou desaparecendo
        if not self.objenabled:
            self.opacity = max(0, self.opacity-255*dt)
            self.position = (self.position[0], self.position[1]+30*dt)
            if self.opacity == 0:
                self.destroy_forever()

# Feather
# pena
class Feather(Collectable):

    # __init__()
    # construtor
    def __init__(self, position, **kwargs):
        super(Feather, self).__init__('feather', position, **kwargs)
        self.timer = 0

    # on_touch()
    # player me pegou
    def on_touch(self):
        p = self.parent.player
        p.set_feathers( p.get_feathers()+1 )
        sound.play_sample('Chip074.wav')

    # update()
    # atualizacao
    def update(self, dt):
        super(Feather, self).update(dt)
        self.timer += dt
        b = 128 + 127*math.cos(2*math.pi * self.timer)
        self.color = (255,255,b)

# Heart
# coracao
class Heart(Collectable):

    # __init__()
    # construtor
    def __init__(self, position, **kwargs):
        super(Heart, self).__init__('heart', position, **kwargs)
        self.position = (self.position[0]-7, self.position[1])

    # on_touch()
    # player me pegou
    def on_touch(self):
        p = self.parent.player
        p.set_hp( p.get_hp() + 30 )
        sound.play_sample('Chip116.wav')


# DeadActor
# usado quando um actor morre...
class DeadActor(Actor):

    # __init__()
    # construtor
    def __init__(self, sender, speed):
        super(DeadActor, self).__init__(sender.animation_name, sender.position, flip_x=sender.flip_x, flip_y=True, anchor=sender.anchor)
        self.speed = (speed[0], max(-200, speed[1]))
        self.i_am_player = (sender.animation_name[:6] == 'player') # gambi, putz...
        self.i_am_boss = (sender.animation_name[:4] == 'boss') # gambi, putz...
        self.color = sender.color

    # update()
    # atualizacao
    def update(self, dt):
        super(DeadActor, self).update(dt)
        self.animation_frame = 0
        (x,y) = self.position
        (sx,sy) = self.speed
        x += sx*dt
        sy += -GRAVITY*dt
        y += sy*dt + 0.5*GRAVITY*dt*dt
        self.speed = (sx,sy)
        self.position = (x,y)

        if y < -self.image.height:
            if self.i_am_player:
                director.replace( FadeTransition(GameOver(), duration=2) )
            elif self.i_am_boss:
                director.replace( FadeTransition(Congratulations(), duration=2) )
            self.destroy()

# =====================================================================


# Projectile
# representa um projetil generico
class Projectile(Actor):

    # __init__()
    # construtor
    def __init__(self, animation_name='miniball', position=(0,0), sender=None, damage=10, rotation=0, scale=1, opacity=255, color=(255,255,255), anchor=None, flip_x=False, flip_y=False, blend_mode = BLEND_STANDARD):
        super(Projectile, self).__init__(animation_name, position, rotation, scale, opacity, color, anchor, flip_x, flip_y, blend_mode)
        self.sender = sender # quem disparou este projetil?
        self.damage = damage # quanto de dano causa?

    # update()
    # atualizacao
    def update(self, dt):
        super(Projectile, self).update(dt)

        # se o tiro sair da tela...
        if not self.inside_screen():
            self.destroy()

    # is_projectile()
    # verifica se e' um projetil
    def is_projectile(self):
        return True

# BirdShot
# tiro do passaro
class BirdShot(Projectile):

    # __init__()
    # construtor
    def __init__(self, position, sender=None, speed=(0,0), hflip=False):
        super(BirdShot, self).__init__('miniball', position, sender, damage=10, flip_x=hflip, anchor=(10,0), blend_mode = BLEND_ADDITIVE)
        self.speed = speed
        self.timer = 0.0

    # update()
    # atualizacao
    def update(self, dt):
        super(BirdShot, self).update(dt)

        # movimentacao
        (x,y) = self.position
        x += self.speed[0]*dt
        y += self.speed[1]*dt
        self.position = (x,y)

        # se o tiro encostar num obstaculo
        if self.parent.collides(self.position):
            self.destroy()

        # cor e dano
        p = self.parent.player
        u = int((255 * p.get_feathers()) / PLAYER_MAXFEATHERS)
        self.color = (u, 255-u, 255-u)
        self.damage = ((50-10)/PLAYER_MAXFEATHERS) * p.get_feathers() + 10

        # timer
        self.timer += dt
        if self.timer >= 0.03:
            self.timer = 0.0
            signal = math.fabs(self.speed[0])/self.speed[0]
            s = (signal*random.randint(0,100), random.randint(-100,100))
            obj = ShotEffect(self.position, s, self.color)
            self.parent.create_actor(obj, zindex=4)

# ShotEffect
# efeito do tiro
class ShotEffect(Projectile):

    def __init__(self, position, speed, col):
        super(ShotEffect, self).__init__('miniball', position, blend_mode=BLEND_ADDITIVE, color=col)
        self.speed = speed

    def update(self, dt):
        super(ShotEffect, self).update(dt)

        eps = 10
        dec = 100 # desaceleracao
        (x,y) = self.position
        (sx,sy) = self.speed
        if (math.fabs(sx) < eps) and (math.fabs(sy) < eps):
            self.destroy()
        else:
            if math.fabs(sx) >= eps:
                signal = math.fabs(sx)/sx
                sx -= signal*dec*dt
            if math.fabs(sy) >= eps:
                signal = math.fabs(sy)/sy
                sy -= signal*dec*dt
        x += sx*dt
        y += sy*dt
        self.speed = (sx,sy)
        self.position = (x,y)

        if self.parent.collides(self.position):
            self.destroy()


# EnemyShot
# tiro do inimigo
class EnemyShot(Projectile):

    # __init__()
    # construtor
    def __init__(self, position, sender=None, speed=(0,0), hflip=False, color=(255,128,0), **kwargs):
        super(EnemyShot, self).__init__('miniball', position, sender, damage=10, flip_x=hflip, anchor=(10,0), blend_mode = BLEND_ADDITIVE, **kwargs)
        self.speed = speed
        self.timer = 0.0
        self.color = color

    # update()
    # atualizacao
    def update(self, dt):
        super(EnemyShot, self).update(dt)

        # movimentacao
        (x,y) = self.position
        x += self.speed[0]*dt
        y += self.speed[1]*dt
        self.position = (x,y)

        # se o tiro encostar num obstaculo
        if self.parent.collides(self.position):
            self.destroy()

        # acertar jogador
        player = self.parent.player
        if self.collides(player) and not(player.is_blinking()):
            player.hit(self.damage)
            self.destroy()

        # timer
        self.timer += dt
        if self.timer >= 0.03:
            self.timer = 0.0
            signal = math.fabs(self.speed[0])/self.speed[0]
            s = (signal*random.randint(0,100), random.randint(-100,100))
            obj = ShotEffect(self.position, s, self.color)
            self.parent.create_actor(obj, zindex=4)

# Rock
# pedra atirada pelo inimigo
class Rock(Projectile):

    # __init__()
    # construtor
    def __init__(self, position, sender, a, target, **kwargs):
        super(Rock, self).__init__('rock', position, sender, damage=10, **kwargs)
        self.flip_x = sender.flip_x

        # calcula coeficientes da parabola y = ax^2 + bx + c
        (xq,yq) = position
        (xp,yp) = target
        self.a = a
        if xq != xp:
            self.b = (yq - yp - self.a*(xq*xq - xp*xp))/(xq-xp)
        else:
            self.b = 10
        self.c = yp - self.a*xp*xp - self.b*xp

        # velocidade
        if position[0] < target[0]:
            self.speed = (400,0)
        else:
            self.speed = (-400,0)

    # update()
    # atualizacao
    def update(self, dt):
        super(Rock, self).update(dt)

        # movimentacao
        (a,b,c) = (self.a, self.b, self.c)
        (x,y) = self.position
        x += self.speed[0]*dt
        y = a*(x*x) + b*x + c
        self.position = (x,y)

        # acertar jogador
        player = self.parent.player
        if self.collides(player) and not(player.is_blinking()):
            player.hit(self.damage)


# =====================================================================

#Objeto sujeito a gravidade
class Fallable(Actor):
    def __init__(self, anim, position, **kwargs):
        super(Fallable, self).__init__(anim, position, **kwargs)
        self.speed = (0, 0)
        self.falling = True
        
    def _check_down_collision(self, point):
        bl = point
        br = (bl[0] + self.width, bl[1])        
        return self.parent.collides(br) or self.parent.collides(bl)
        
    def _check_up_collision(self, point):
        bl = point
        tl = (bl[0], bl[1] + self.height)
        tr = (bl[0] + self.width, bl[1] + self.height)
        return self.parent.collides(tl) or self.parent.collides(tr)
        
    def _check_left_collision(self, point):
        bl = point
        tl = (bl[0], bl[1] + self.height)
        return self.parent.collides(bl) or self.parent.collides(tl)

    def _check_right_collision(self, point):
        bl = point
        br = (bl[0] + self.width, bl[1])        
        tr = (bl[0] + self.width, bl[1] + self.height)
        return self.parent.collides(br) or self.parent.collides(tr)

    def _check_collision(self, point):
        return self._check_down_collision(point) or self._check_up_collision(point)
    
    def _fall(self, dt):
        sx, sy = self.speed
        x, y = self.position
        sy += -GRAVITY*dt
        y += sy*dt + 0.5*GRAVITY*dt*dt
        return (x, y), (sx, sy)
        
    def jump(self, sy):
        self.speed = (self.speed[0], sy)

    def walk(self, dt, direction):
        x, y = self.position
        x += direction * self.speed[0] * dt
        if not self._check_collision((x, y)):
            self.position = x, y
        else:
            pass

    def _snap_down(self, point):
        x, y = point
        mx, my = self.parent.get_index((x,y))
        my = self.parent.height - my * self.parent.tile_height
        newy = my
        self.position = x, newy

    def _snap_up(self, point):
        x, y = point
        mx, my = self.parent.get_index((x,y))
        my = self.parent.height - my * self.parent.tile_height
        newy = my - self.parent.tile_height
        self.position = x, newy

    def fall(self, dt):
        if dt > 0.1: return
        (x, y), (sx, sy) = self._fall(dt)
        if not self._check_down_collision((x, y)):
            self.falling = True
            self.position = x, y
            self.speed = sx, sy
        else:
            if self.falling:
                self.falling = False
                self.speed = (sx, 0)
                self._snap_down((x,y))
        if self._check_up_collision((x, y)):
            self.speed = (self.speed[0], -0.01)
            self._snap_up((x,y))

    def update(self, dt):
        if self.speed[1] < -500: self.speed = (self.speed[0], -500)
        super(Fallable, self).update(dt)
        if dt > 0.08: return
        self.fall(dt)
        
    def is_falling(self):
        return not self.speed[1] == 0 or self.falling
                

# =====================================================================

# Spikes
# espinhos
class Spikes(Actor):

    # __init__()
    # construtor
    def __init__(self, position):
        super(Spikes, self).__init__('spikes', position, anchor=(0,32))

    # update()
    # atualizacao
    def update(self, dt):
        p = self.parent.player
        if self.collides(p):
            p.hit(30)

# =====================================================================

# Enemy
# inimigo generico
class Enemy(Fallable):

    # __init__()
    # construtor
    def __init__(self, animation, position, speed_x=100, jump_strength=400, damage=34, **kwargs):
        super(Enemy, self).__init__(animation, position, **kwargs)
        self.dead = False
        self.jump_strength = jump_strength
        self.damage = damage # quanto de dano causa ao player
        self.input = set()
        self.speed = (speed_x, 0)

    # update()
    # atualiza inimigo
    def update(self, dt):
        if self.dead: return
        super(Enemy, self).update(dt)

        # movimentacao
        (sx,sy) = self.speed
        if (LEFT in self.input) and not(RIGHT in self.input):
            self.walk(dt, -1)
            self.flip_x = True
        elif (RIGHT in self.input) and not (LEFT in self.input):
            self.walk(dt, 1)
            self.flip_x = False
        if UP in self.input and not self.is_falling():
            sy = self.jump_strength
        self.speed = (sx,sy)

        try:
            player = self.parent.player

            # fui atingido!
            projectiles = self.parent.get_projectiles()
            for p in projectiles:
                if self.collides(p) and (p.sender == player) and not(self.is_blinking()):
                    self.hit(p.damage)
                    p.destroy()

            # se encostar no jogador
            if self.collides(player):
                player.hit(self.damage)
        except:
            pass

# Snake
# cobrinha... =P
class Snake(Enemy):

    # __init__()
    # construtor
    def __init__(self, position):
        super(Snake, self).__init__('snake', position, damage=20, hp=10)
        self.input.add(RIGHT)

    # update()
    # atualizacao automatica
    def update(self, dt):

        # movimentacao <-- e -->
        (tl, t, tr, r, br, b, bl, l) = self.get_collision_detectors((2,5))
        if l or (not(bl) and br):
            self.input.clear()
            self.input.add(RIGHT)
        if r or (not(br) and bl):
            self.input.clear()
            self.input.add(LEFT)

        # atualizacao
        super(Snake, self).update(dt)

# RedSnake
# cobra + forte
class RedSnake(Enemy):

    # __init__()
    # construtor
    def __init__(self, position):
        super(RedSnake, self).__init__('snake', position, damage=20, hp=20, color=(192,0,32))
        self.input.add(RIGHT)

    # update()
    # atualizacao automatica
    def update(self, dt):

        # movimentacao <-- e -->
        (tl, t, tr, r, br, b, bl, l) = self.get_collision_detectors((2,5))
        if l or (not(bl) and br):
            self.input.clear()
            self.input.add(RIGHT)
        if r or (not(br) and bl):
            self.input.clear()
            self.input.add(LEFT)

        # atualizacao
        super(RedSnake, self).update(dt)


# RockGuy
# cara que atira pedras
class RockGuy(Enemy):

    # __init__()
    # construtor
    def __init__(self, position):
        super(RockGuy, self).__init__('rockguy', position, damage=20, hp=50)
        self._ok = True

    # update()
    # atualizacao automatica
    def update(self, dt):

        # onde esta o jogador?
        (x,y) = self.position
        (xp,yp) = self.parent.player.position
        self.flip_x = (x > xp)

        # atira pedras
        if int(self.animation_frame) == 1 and self._ok and self.inside_screen():
            self._ok = False
            pos = (self.position[0], self.position[1] + self.image.height*0.4)
            sound.play_sample('Chip032.wav')
            if abs(x-xp) >= 50:
                t = (xp,yp)
            elif self.flip_x:
                t = (x-100,y+50)
            else:
                t = (x+100,y+50)
            self.parent.create_actor( Rock(pos, sender=self, a=-0.01, target=t), zindex=4 )
        if int(self.animation_frame) != 1:
            self._ok = True

        # atualizacao
        super(RockGuy, self).update(dt)


# Monster
# monstro grande
class Monster(Enemy):

    # __init__()
    # construtor
    def __init__(self, position):
        super(Monster, self).__init__('monster', position, damage=10, hp=40, jump_strength=500)
        self.speed = (70, self.speed[1])

    # update()
    # atualizacao automatica
    def update(self, dt):
        (tl, t, tr, r, br, b, bl, l) = self.get_collision_detectors((5,2))
        (x, y) = self.position
        (px, py) = self.parent.player.position
        eps = 5

        self.input.clear()

        if x < px-eps:
            self.input.add(RIGHT)
        elif x > px+eps:
            self.input.add(LEFT)

        if (r or l) and not self.is_falling():
            self.input.add(UP)
            sound.play_sample('Chip103.wav')

        super(Monster, self).update(dt)

# Skull
# caveira
class Skull(Enemy):

    # __init__()
    # construtor
    def __init__(self, position):
        super(Skull, self).__init__('skull', position, damage=30, hp=10)

    # update()
    # atualizacao
    def update(self, dt):
        (x, y) = self.position
        (px, py) = self.parent.player.position
        eps = 5

        self.input.clear()
        self.input.add(UP)
        if x < px-eps:
            self.input.add(RIGHT)
        elif x > px+eps:
            self.input.add(LEFT)

        super(Skull, self).update(dt)

# Drunk
# monstro bebado
class Drunk(Enemy):

    # __init__()
    # construtor
    def __init__(self, position):
        super(Drunk, self).__init__('drunk', position, damage=20, hp=50)
        self.input.add(RIGHT)
        self.set_mp(0)

    # update()
    # atualizacao automatica
    def update(self, dt):
        super(Drunk, self).update(dt)

        # movimentacao
        (tl, t, tr, r, br, b, bl, l) = self.get_collision_detectors((2,5))
        (x, y) = self.position
        (px, py) = self.parent.player.position
        (dx, dy) = (px-x, py-y)
        magnitude = math.sqrt(dx*dx+dy*dy)
        eps = 5

        if magnitude < 200:
            # andando
            self.input.clear()
            if x < px-eps and br:
                self.input.add(RIGHT)
            elif x > px+eps and bl:
                self.input.add(LEFT)
            else:
                self.animation_frame = 0
            if (l or r) and not self.is_falling():
                self.input.add(UP)
                sound.play_sample('Chip103.wav')


            # atirando
            self.set_mp( self.get_mp() + (self.get_maxmp()*0.5)*dt )
            if self.get_mp() >= self.get_maxmp() and self.inside_screen():
                self.set_mp(0)
                sound.play_sample('Chip054.wav')
                try:

                    if self.flip_x:
                        pos = (self.position[0], self.position[1] + self.image.height*0.4)
                        self.parent.create_actor( EnemyShot(pos, sender=self, speed=(-500, 0), hflip=True), zindex=4 )
                    else:
                        pos = (self.position[0] + self.image.width, self.position[1] + self.image.height*0.4)
                        self.parent.create_actor( EnemyShot(pos, sender=self, speed=(500, 0), hflip=False), zindex=4 )

                except:
                    print 'erro ao atirar: ', sys.exc_info()
        else:
            # andando
            if l or (not(bl) and br):
                self.input.clear()
                self.input.add(RIGHT)
            if r or (not(br) and bl):
                self.input.clear()
                self.input.add(LEFT)


# Boss
# chefao
class Boss(Enemy):

    # __init__()
    # construtor
    def __init__(self, position):
        super(Boss, self).__init__('boss', position, damage=10, hp=500, jump_strength=500)
        self.speed = (200, self.speed[1])

    # update()
    # atualizacao automatica
    def update(self, dt):
        # movimentacao
        (tl, t, tr, r, br, b, bl, l) = self.get_collision_detectors((5,2))
        (x, y) = self.position
        (px, py) = self.parent.player.position
        eps = 5





        # atirando
        self.set_mp( self.get_mp() + (self.get_maxmp()*0.2)*dt )
        if self.get_mp() >= self.get_maxmp() and self.inside_screen():
            self.set_mp(0)
            self.input.clear()
            sound.play_sample('Chip054.wav')
            if x < px-eps:
                self.input.add(RIGHT)
            elif x > px+eps:
                self.input.add(LEFT)
            try:
                pos = (self.position[0] + self.image.width*0.5, self.position[1] + self.image.height*0.4)
                v = 320
                col = (0, 128, 255)
                self.parent.create_actor( EnemyShot(pos, sender=self, speed=(-v, 0), hflip=True, color=col), zindex=4)
                self.parent.create_actor( EnemyShot(pos, sender=self, speed=(-v, -v), hflip=True, color=col), zindex=4)
                self.parent.create_actor( EnemyShot(pos, sender=self, speed=(v, 0), hflip=True, color=col), zindex=4)
                self.parent.create_actor( EnemyShot(pos, sender=self, speed=(v, -v), hflip=True, color=col), zindex=4)
            except:
                print 'erro ao atirar: ', sys.exc_info()

        # pulo
        if UP in self.input:
            self.input.remove(UP)
        if b and (random.randint(0, 100) <= 3) and not self.is_falling():
            sound.play_sample('Chip103.wav')
            self.input.add(UP)

        # atualizacao
        super(Boss, self).update(dt)


# =====================================================================

# Player
# jogador
class Player(Fallable):
    
    # __init__()
    # construtor
    def __init__(self, position, **kwargs):
        super(Player, self).__init__('player_stopped', position, **kwargs)
        self.speed = (200,0)
        self.dead = False
        self._hp = player_hp
        self._mp = player_mp
        self._maxhp = PLAYER_MAXHP
        self._maxmp = PLAYER_MAXMP

    # set_hp()
    # define novo hp
    def set_hp(self, x):
        global player_hp
        super(Player, self).set_hp(x)
        player_hp = self.get_hp()


    # set_mp()
    # define novo mp
    def set_mp(self, x):
        global player_mp
        super(Player, self).set_mp(x)
        player_mp = self.get_mp()

    # set_feathers()
    # define nova qtd de penas
    def set_feathers(self, x):
        global player_feathers
        player_feathers = max(0, min(PLAYER_MAXFEATHERS, x))
        if player_feathers == PLAYER_MAXFEATHERS:
            self.parent.change_level(FINAL_LEVEL[0], (FINAL_LEVEL[1][0], FINAL_LEVEL[1][1]))

    # get_feathers()
    # retorna qtd de penas
    def get_feathers(self):
        return player_feathers


    # update()
    # atualiza player (chamado automaticamente)
    def update(self, dt):
        super(Player, self).update(dt)
        input = self.parent.chars_pressed
        (tl, t, tr, r, br, b, bl, l) = self.get_collision_detectors((2,5))
        
        # aumenta mp
        self.set_mp( self.get_mp() + 30*dt )

        # movimentacao no eixo x
        (x,y) = self.position
        (sx,sy) = self.speed
        if (LEFT in input) and not(RIGHT in input):
            self.walk(dt, -1)
            if b: self.change_animation('player_walking')
            else: self.change_animation('player_jumping')
            self.flip_x = True
        elif (RIGHT in input) and not (LEFT in input):
            self.walk(dt, 1)
            if b: self.change_animation('player_walking')
            else: self.change_animation('player_jumping')
            self.flip_x = False
        else:
            if b: self.change_animation('player_stopped')
            else: self.change_animation('player_jumping')

        # movimentacao no eixo y
        if UP in input and not self.is_falling():
            sy = 400
            sound.play_sample('Chip103.wav')
        self.speed = (sx,sy)        

        # transicao de fase
        tile = self.get_my_tile()
        if tile[0] == SPECIAL:
            data = tile[1]
            self.parent.change_level(data[0], (data[1], data[2]))


# Bird
# Passaro
class Bird(Actor):

    # __init__()
    # construtor
    def __init__(self, position):
        super(Bird, self).__init__('bird', position)
        self.speed = (150,150)


    # update()
    # atualiza player (chamado automaticamente)
    def update(self, dt):
        super(Bird, self).update(dt)
        try:
            p = self.parent.player
            if (p is None) or p.is_dead():
                return

            # flip
            self.flip_x = p.flip_x

            # movimentacao
            if p.flip_x: eps = 40
            else: eps = -40
            (dx, dy) = ((p.position[0]+eps)-self.position[0], (p.position[1]+10)-self.position[1])
            magnitude = math.sqrt(dx*dx + dy*dy)
            (dx, dy) = (dx/magnitude, dy/magnitude)
            if magnitude > 5.0:
                (sx,sy) = self.speed
                (x,y) = self.position
                x += dx*sx*dt
                y += dy*sy*dt
                self.position = (x,y)

            # atirar
            input = self.parent.chars_pressed
            if (LEFT_CTRL in input) and (p.get_mp() == p.get_maxmp()) and self.inside_screen():
                input.remove(LEFT_CTRL)
                sound.play_sample('Chip054.wav')
                p.set_mp(0)
                if self.flip_x:
                    pos = (self.position[0], self.position[1] + self.image.height*0.4)
                    self.parent.create_actor( BirdShot(pos, sender=p, speed=(-500, 0), hflip=True), zindex=4 )
                else:
                    pos = (self.position[0] + self.image.width, self.position[1] + self.image.height*0.4)
                    self.parent.create_actor( BirdShot(pos, sender=p, speed=(500, 0), hflip=False), zindex=4 )
        except:
            pass
