# --------------------------------
# maploader.py
# Carregador de mapas
# --------------------------------
import sys
import string
import data
from TiledLayer import *
from actors import *


# MapLoader
# carrega mapa de arquivo
class MapLoader:

    # __init__()
    # construtor
    def __init__(self, filename):
        self._load_from_file(filename)

    # get_dimensions()
    # retorna (largura,altura) do mapa (em tiles)
    def get_dimensions(self):
        return (len(self.data[0]), len(self.data))

    # export_to()
    # exporta mapa para uma TiledLayer
    def export_to(self, layer):
        (w,h) = self.get_dimensions()

        try:
            bg = pyglet.resource.image(data.directory(self.background))
            layer.set_bg( Sprite(bg,position=(0,0)) )
        except:
            pass

        for r in range(0, h):
            for c in range(0, w):
                layer[r,c] = self.data[r][c]

        for e in self.object_list: # e = (nome, (xpos,ypos))
            obj = None
            pos = (e[1][0] * layer.tile_width, layer.height - e[1][1] * layer.tile_height)

            if e[0] == 'spikes': (obj, z) = (Spikes(pos), 2)
            elif e[0] == 'feather': (obj, z) = (Feather(pos), 4)
            elif e[0] == 'heart': (obj, z) = (Heart(pos), 4)
            elif e[0] == 'snake': (obj, z) = (Snake(pos), 3)
            elif e[0] == 'redsnake': (obj, z) = (RedSnake(pos), 3)
            elif e[0] == 'drunk': (obj, z) = (Drunk(pos), 3)
            elif e[0] == 'skull': (obj, z) = (Skull(pos), 3)
            elif e[0] == 'rockguy': (obj, z) = (RockGuy(pos), 3)
            elif e[0] == 'monster': (obj, z) = (Monster(pos), 3)
            elif e[0] == 'boss': (obj, z) = (Boss(pos), 3)
            else: print 'unknown object: %s' % (e[0])

            if obj != None:
                layer.create_actor(obj, z)

    # _load_from_file()
    # carrega mapa de arquivo
    def _load_from_file(self, filename):
        self.name = 'Noname'
        self.tileset = ''
        self.music = ''
        self.background = ''
        self.table = { }
        self.data = []
        self.object_list = []

        area = ''
        lc = 0 # contador de linha
        fp = data.load(filename, 'r')
        lines = fp.readlines()
        for r in lines:
            r = string.strip(r)
            if r == '': continue

            # muda secao
            if r in ('[header]', '[table]', '[data]'):
                area = r
                lc = 0
                continue

            # qual secao?
            if area == '[header]':
                if lc == 0: self.name = r
                elif lc == 1: self.tileset = r
                elif lc == 2: self.music = r
                elif lc == 3: self.background = r
            elif area == '[table]':
                s = string.split(r, ' ')
                if s[1] == 'wall':
                    self.table[s[0]] = ( int(s[2]) , BLOCK , None )
                elif s[1] == 'passable':
                    self.table[s[0]] = ( int(s[2]) , PASS , None )
                elif s[1] == 'warp':
                    self.table[s[0]] = ( 0 , SPECIAL , (s[2], int(s[3]), int(s[4])) )
                elif s[1] == 'object':
                    self.table[s[0]] = s[2]
            elif area == '[data]':
                x = []
                cc = 0 # contador de coluna
                for c in r:
                    try:
                        k = self.table[c]
                        if type(k) != type('string'):
                            x.append(k)
                        else:
                            x.append( (0, PASS, None) )
                            self.object_list.append( (k, (cc, lc)) )
                    except:
                        print 'unknown block symbol (%s): %s' % (filename, c)
                    cc += 1

                self.data.append(x)

            lc += 1

        fp.close()


