# --------------------------------
# testlevel.py
# Fase de testes
# --------------------------------
from cocos.director import *
from cocos.scene import *
from cocos.scenes import *
from cocos.layer import *
from cocos.sprite import *
from cocos.text import *
from gamedata import *
from TiledLayer import *
from Level import *
from maploader import *
import pyglet
import math
import random
import sound
import sys



levelname = ''

# TitleLabel
# label que serve como titulo
class TitleLabel(Label):

    # __init__()
    # construtor
    def __init__(self, **kwargs):
        self.timer = 0
        super(TitleLabel, self).__init__(**kwargs)

    # update()
    # atualizacao
    def update(self, dt):
        self.timer += dt
        if self.timer > 3:
            self.opacity = max(0, 255 - 128*(self.timer-3))


# TestLevelForeground
# layer da fase de testes
class TestLevelForeground(Level):

    is_event_handler = True

    # __init__()
    # construtor
    def __init__(self, tileset, player_position, width, height, mapfile):
        self.chars_pressed = set()
        self._active = True
        #print data.filepath(tileset)
        img = pyglet.image.load( data.filepath(tileset) ) # TODO: evitar carregar a imagem 2 vezes
        (cols, rows) = (img.width / TILE_WIDTH , img.height / TILE_HEIGHT)
        super(TestLevelForeground, self).__init__((tileset, rows, cols), width, height, player_position)
        self.mapfile = mapfile

    # on_key_press()
    # evento: tecla pressionada?
    def on_key_press(self, key, modifiers):
        self.chars_pressed.add(key)

    # on_key_release()
    # evento: tecla solta?
    def on_key_release(self, key, modifiers):
        if key in self.chars_pressed:
            self.chars_pressed.remove(key)

    # create_actor()
    # cria um sprite, definindo actor.parent = esta layer
    def create_actor(self, actor, zindex=1):
        self.add(actor, z=zindex)

    # update()
    # atualizacao da layer
    def update(self, dt):
        if not self._active: return
        super(TestLevelForeground, self).update(dt)

        # atualiza actors
        for c in self.children:
            try:
                c[1].update(dt)
                if c[1].is_dead():
                    self.remove(c[1])
            except: # nao e' actor
                pass

    # get_projectiles
    # devolve o conjunto de projeteis ativos
    def get_projectiles(self):
        x = []
        for c in self.children:
            try:
                if c[1].is_projectile():
                    x.append(c[1])
            except:
                pass
        return x

    # change_level()
    # muda o mapa
    def change_level(self, mapname, position):
        #print 'changing level... ', mapname, position
        try:
            lev = TestLevel(mapname, position)
            director.replace( FadeTransition(lev, duration=2) )
            self._active = False
        except:
            print 'error: ', sys.exc_info()

# Hud
# layer da hud
class Hud(Layer):

    def __init__(self, gamelayer):
        super(Hud, self).__init__()

        self.gamelayer = gamelayer
        self.timer = 0.0

        self.lbl_hp = Label('Energy: ?%', font_name=FONT_NAME, font_size=14, x=5, y=10)
        self.lbl_hp2 = Label('Energy: ?%', font_name=FONT_NAME, font_size=14, x=5+1, y=10-1, color=(0,0,0,255))
        self.add(self.lbl_hp, z=1)
        self.add(self.lbl_hp2)
        self.lbl_mp = Label('Magic: ?%', font_name=FONT_NAME, font_size=14, x=SCREEN_WIDTH/2, y=10, anchor_x='center')
        self.lbl_mp2 = Label('Magic: ?%', font_name=FONT_NAME, font_size=14, x=SCREEN_WIDTH/2+1, y=10-1, anchor_x='center', color=(0,0,0,255))
        self.add(self.lbl_mp, z=1)
        self.add(self.lbl_mp2)
        self.lbl_feathers = Label('x ? of ?', font_name=FONT_NAME, font_size=14, x=SCREEN_WIDTH-85, y=10)
        self.lbl_feathers2 = Label('x ? of ?', font_name=FONT_NAME, font_size=14, x=SCREEN_WIDTH-85+1, y=10-1, color=(0,0,0,255))
        self.add(self.lbl_feathers, z=1)
        self.add(self.lbl_feathers2)

        self.feather = Sprite(data.directory('feather.png'), position=(SCREEN_WIDTH-105, 30), scale=2)
        self.add(self.feather, z=1)

        self.unschedule(self.update)
        self.schedule_interval(self.update, 1.0/FPS)

    def update(self, dt):
        global game_time
        self.timer += dt
        game_time += dt

        for c in self.children:
            try:
                c[1].update(dt)
            except:
                pass

        try:
            player = self.gamelayer.player

            self.lbl_hp.element.text = 'Energy: %d%%' % (int(100*player.get_hp() / player.get_maxhp()))
            g = int((255 * player.get_hp()) / player.get_maxhp())
            self.lbl_hp.element.color = (255-g,g,0,255)

            self.lbl_mp.element.text = 'Magic: %d%%' % (int(100*player.get_mp() / player.get_maxmp()))
            g = int((255 * player.get_mp()) / player.get_maxmp())
            self.lbl_mp.element.color = (255-g,g,0,255)

            self.lbl_feathers.element.text = 'x %d of %d' % (player.get_feathers(), PLAYER_MAXFEATHERS)

            self.lbl_hp2.element.text = self.lbl_hp.element.text
            self.lbl_mp2.element.text = self.lbl_mp.element.text
            self.lbl_feathers2.element.text = self.lbl_feathers.element.text

            t = 127*math.cos((0.02*math.pi)*game_time) + 128
            self.feather.color = (255,255,t)
        except:
            pass


# TestLevel
# zona de testes
class TestLevel(Scene):

    # __init__()
    # construtor
    def __init__(self, mapfile, player_position):
        self.maploader = MapLoader(mapfile)
        (w, h) = self.maploader.get_dimensions()
        (w, h) = (w * TILE_WIDTH, h * TILE_HEIGHT)
        #h = 512
        bg = None
        fg = TestLevelForeground(self.maploader.tileset, player_position, w, h, mapfile)
        hud = Hud(fg)
        self.maploader.export_to( fg )
        fg.generate()
        global levelname
        if levelname != self.maploader.name:
            levelname = self.maploader.name
            title = TitleLabel(text=levelname, font_name=FONT_NAME, font_size=32, x=SCREEN_WIDTH/2, y=SCREEN_HEIGHT/2, anchor_x='center', anchor_y='center')
            hud.add(title, z=1)


        super(TestLevel, self).__init__(fg, hud)
        sound.play_music(self.maploader.music)


